#include <iostream>
#include <fstream>

#include <core/image/ImageT.h>
#include <core/basics/Config.h>

#include <vislearning/baselib/Globals.h>
#include <vislearning/cbaselib/MultiDataset.h>
#include <vislearning/features/localfeatures/LFonHSG.h>
#include <vislearning/features/localfeatures/LFWriteCache.h>

using namespace std;
using namespace NICE;
using namespace OBJREC;

int main(int argc, char **argv)
{
  Config conf ( argc, argv );

  string destForFeat = conf.gS("main", "destForFeat", "/tmp/");
  bool verbose = conf.gB("main", "verbose", false);

  MultiDataset md ( &conf );

  const LabeledSet & all = *(md["all"]);

  const ClassNames & classNames = md.getClassNames("all");

  LocalFeatureRepresentation *lfrep = NULL;
  lfrep = new LFonHSG(&conf);
  
  LocalFeatureRepresentation *lfrep_save = new LFWriteCache ( &conf, lfrep );
  lfrep = lfrep_save;
  
  LOOP_ALL_S ( all ) 
  {
    EACH_S(classno, imgfilename);
    Globals::setCurrentImgFN ( imgfilename ); 
    if (verbose)
      std::cerr << "imgfilename: " << imgfilename << std::endl;
    Image img (imgfilename);
    NICE::VVector positions;
    NICE::VVector features;
    // Extract the descriptor-Values from a given grayscale-Image.
    lfrep->extractFeatures ( img, features, positions );
    if (verbose)
      std::cerr << "features.size: " << features.size() << std::endl;
  }

  return 0;
}
